//    (c)2k7 Michal Glebowski, michal@glebowski.pl
//    http://glebowski.pl/blog/delphi/tsymlabel/
//    freeware, MIT License
//
//    The component faciliates entering math equations,
//    superscripts, subscripts, greek letters, sums, integrals, fractions.
//    It is an descendant of Tlabel.
//
//    tested on Turbo Delphi
//
//    samples: y = ax^2 + bx + c_1 ,  D_1, Q_t, Q_n, Q_{min}
//             #D = b^2 - 4 a c
//
//   #Dp_k = ##fra{#Dp_L;K} = #Dp_1
//   Q_N - #DQ_{N2dop} #< Q_j #< Q_N + #DQ_{N1dop}
//   p_{at} = p_0 = 1,013 bar ##arr K_p = 1 + k_w  p_{sr}
//   t_{sr} = ##fra{1;#t_r} ##int{0; #t_r} t_d(#t) #@ ##fra{1;#t_r} ##sum{r=1;R} t_d(#t_r)#D#t_r

//
//    beware: font height must be expressed in negative values (for now)!!!
//

unit SymLabel;

interface

uses
  SysUtils, Classes, Controls, StdCtrls, Messages, Windows, Graphics;

type
  TSymLabel = class(TCustomLabel)
  private
    function DrawParse(x, BaseY: Integer; txt: String): Integer;  
  protected { Protected declarations }
    procedure DoDrawText(var Rect: TRect; Flags: Longint); override;
  public { Public declarations }
    constructor Create (AOwner : TComponent); override;
  published { Published declarations }
    Property Align;
    Property Alignment;
    Property AutoSize;
    Property Color;
    Property Cursor;
    Property Caption;
    Property Font;
    Property ParentFont;
  end;

procedure Register;

implementation

uses Types, Math, StrUtils;

procedure Register;
begin
  RegisterComponents('User', [TSymLabel]);
end;

constructor TSymLabel.Create(AOwner: TComponent);
begin
  inherited ;
  AutoSize:= False;
end;

function TSymLabel.DrawParse(x, BaseY: Integer; txt: String): Integer;

    //
    // returns String which width is more accurate to real width
    // due to ommitting special (non-printable) chars
    //
    function RealTxt(s: String): String;

        function CharsReplace(const S, OldPattern, NewPattern: String ): String;
        var i: integer;
        begin
          result:= S;
          if Length(OldPattern)>0 then
            for i := 1 to length(OldPattern) do
              result:= StringReplace(result, OldPattern[i], NewPattern, [rfReplaceAll]);
        end;

    begin
      result:= CharsReplace(s, '_#{}', '');
    end;

    function MatchingCurl(s: String; from: Integer): Integer;
    var p1, p2, n, i: Integer;
    begin
      result:= 0;
      p1:= POSEX('{', s, from);
      p2:= POSEX('}', s, from);
      if (p1>0) and (p2>p1) then begin
        n:= 0;
        for i := p1 to length(s) do
          if s[i]='{' then inc(n)
          else if s[i]='}' then begin
            dec(n);
            if n=0 then begin
              result:= i;
              break;
            end;
          end;
      end;
    end;

var
    y: Integer;
    FHeight: Integer;
    DFont: TFont; // Default Font
    Mode,         // 0 - normal, 1 - subscript, 2 - superscript
    Style,        // 0 - normal, 1 - greek
    Cont: Byte;   // continous mode, 0 - single char, 1 - string (of changed style/mode)
    s: Char;
    i: Integer;
    Token, TokenParam: String;
    OldFHeight: Integer;
    nom, rt1, rt2, par1, par2: String;  tw1, tw2, twn: Integer;
    posSpc, posCu1, posCu2, posSep: Integer;

begin

  Canvas.Font := Font;
  DFont:= TFont.Create;
  DFont.Assign(Canvas.Font);
  FHeight:= ABS(Font.Height);

  i:=1; Mode:= 0; Style:= 0; Cont:= 0;
  while i <= Length(txt) do with Canvas do begin

    s:= txt[i];

    if (s='#') and (i<Length(txt)) and (txt[i+1] in ['(',')', '<', '>']) then begin
        token:= txt[i+1];

        if token='(' then begin
          Font:=DFont;
          Font.Name:= 'Symbol';
          TextOut( x, BaseY - FHeight div 2, #236 ); // gra ( 230
          TextOut( x, BaseY + FHeight div 2, #238 ); // d  ( 232
          inc(x, TextWidth('m'));
        end; // if token='('

        if token=')' then begin
          Font:=DFont;
          Font.Name:= 'Symbol';
          TextOut( x, BaseY - FHeight div 2, #252 ); // gra ) 246
          TextOut( x, BaseY + FHeight div 2, #254 ); // d  ) 248
          inc(x, TextWidth('m'));
        end; // if token='('

        if token='<' then begin
          Font:=DFont;
          Font.Name:= 'Symbol';
          TextOut( x, BaseY , #163 ); // <=
          inc(x, TextWidth('mm'));
        end;

        if token='>' then begin
          Font:=DFont;
          Font.Name:= 'Symbol';
          TextOut( x, BaseY , #179 ); // >=
          inc(x, TextWidth('mm'));
        end;

        inc(i, 2+length(Token));
        continue;

    end else

    if (s='#') and (i<Length(txt)) and (txt[i+1]='#') then begin
        posSpc:= POSEX(#32, txt, i+1);
        posCu1:= POSEX('{', txt, i+1);
        //posCu2:= POSEX('}', txt, i+1);    // may not match ( if some {} inside )
        posCu2:= MatchingCurl(txt, posCu1);

        if (posCu1>0) and
         ( ((posCu1<posSpc) and (posSpc>0)) or (posSpc=0)) then begin

                           Token:= copy(txt, i+2, posCu1 - i - 2);
                      TokenParam:= copy(txt, posCu1+1, posCu2-posCu1-1);

                      posSep:= POS(';', TokenParam);
                      if posSep=0 then posSep:= length(TokenParam);

                      par1:= copy(TokenParam, 1, posSep-1);
                      par2:= copy(TokenParam, posSep+1, length(TokenParam)-posSep);
        end else begin
          TokenParam:= '';
          if posSpc>0 then Token:= copy(txt, i+2, posSpc - i - 2)
                      else Token:= copy(txt, i+2, length(txt));
        end;

        if token='arl' then begin
          Font:=DFont; Font.Name:= 'Symbol';
          TextOut( x, BaseY, #172 ); // <-
          inc(x, TextWidth('mmm'));
        end; // if token='arl'
        if token='arr' then begin
          Font:=DFont; Font.Name:= 'Symbol';
          TextOut( x, BaseY, #174 ); // ->
          inc(x, TextWidth('mmm'));
        end; // if token='arr'

        if token='int' then begin

          Font:=DFont;
          Font.Name:= 'Symbol';
          TextOut( x, BaseY - FHeight div 2, #236 ); // gra caki
          TextOut( x, BaseY + FHeight div 2, #254 ); // d  caki

          if length(TokenParam)>0 then begin
            tw2:= TextWidth(RealTxt(par2)) div 2 - TextWidth('i')*2;
            tw1:= TextWidth(RealTxt(par1)) div 2 + TextWidth('');
            DrawParse(x - tw2, BaseY - Round(FHeight * 1.5), par2);  // gra
            DrawParse(x - tw1, BaseY + Round(FHeight * 1.5), par1);  // d
          end;
          inc(x, TextWidth('m'));

        end; // if token='int'

        if token='sum' then begin
          Font:=DFont;
          Font.Name:= 'Symbol';
          OldFHeight:= Font.Height;
          Font.Height:= Round(Font.Height * 1.8);
          TextOut( x, BaseY - Round(FHeight * 0.5) - 1, #229 ); // suma
          Font.Height:= OldFHeight;

          if length(TokenParam)>0 then begin
            tw2:= TextWidth(RealTxt(par2)) div 2 - TextWidth('m') + 2;
            tw1:= TextWidth(RealTxt(par1)) div 2 - TextWidth('m') + 2;
            DrawParse(x - tw2 + 4, BaseY - Round(FHeight * 1.5), par2);  // gra
            DrawParse(x - tw1 + 4, BaseY + Round(FHeight * 1.5), par1);  // d
          end;
          inc(x, TextWidth('mm') + 4);
        end; // if token='sum'

        if token='fra' then begin

          rt1:= RealTxt(par1);
          rt2:= RealTxt(par2);
          tw1:= TextWidth(rt1) div 2;
          tw2:= TextWidth(rt2) div 2;

          Font:=DFont;
          Font.Name:= 'Symbol';

          twn:= Round(max(tw1, tw2) / TextWidth(#190) * 2) ;
          nom:= StringOfChar(#190, twn);
          TextOut( x, BaseY, nom );
          twn:= TextWidth(nom) div 2;

          if length(TokenParam)>0 then begin
            DrawParse(x + twn - tw1, BaseY - Round(FHeight) div 2 -2, par1);  // gra
            DrawParse(x + twn - tw2, BaseY + Round(FHeight) div 2 +2, par2);  // d
          end;

          inc(x, (twn+1) * 2);
        end; // if token='fra'

        inc(i, 2+length(Token));
        if Length(TokenParam)>0 then inc(i, 2+length(TokenParam));

    end else

      case s of
          '^': Mode:=  1; // superscripts
          '_': Mode:=  2; // subscripts
          '#': Style:= 1; // greek letters
          '{': Cont:=  1; // sets (more than one letter)
          '}': begin Mode:= 0; Style:= 0; Cont:= 0; end;
          else begin

            Font:=DFont;
            if Style=1 then Font.Name:= 'Symbol';

            case Mode of
              1:   y:= BaseY - FHeight div 3 - 3 + floor(FHeight/3);
              2:   y:= BaseY + FHeight div 3 + 1 + floor(FHeight/6);
              else y:= BaseY;
            end;

            if Mode>0 then begin //Canvas.Font.Height:=  * 9 div 10;
               Font.Size:= round (Font.Size / 1.6);
               if Font.Size<7 then Font.Size:= 7;
            end;

            TextOut( x, y, s );
            inc(x, TextWidth(s));

            if Cont = 0 then begin
              Mode:= 0;
              Style:= 0;
            end;
          end;

      end; // case s
      inc(i);

  end; // while

  result:= x;

  DFont.Free;
end;

procedure TSymLabel.DoDrawText(var Rect: TRect; Flags: Longint);
var FHeight: Integer;
    RealHeight,
    BaseY: Integer;
begin

  Canvas.Font := Font;

  FHeight:= ABS(Font.Height);
  RealHeight:= Round(1.22 * FHeight - 0.15);
  BaseY:= (Self.Height - RealHeight) div 2 - 1;

  Rect.Right := DrawParse(0, BaseY, Text);

end;



{$region 'Przeliczanie FontHeight (organoleptycznie)'}
{
  NegativeFH = 0,82 * PositiveFH + 0,14
  PositiveFH = 1,22 * NegativeFH - 0,15
}
{$endregion}

end.
